/*
   Parallax Mapping Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to perform parallax mapping in Direct3D.  Parallax mapping
   is simple to implement and is not expensive.  Parallax mapping is a way to simulate
   detail on a surface that basically builds off of bump mapping.  Parallax mapping
   can show detail on a surface that seems to change with the movement of the camera.
   For example if you bumped mapped a wall with a brick normal map you can not see
   detail such as underneath each brick when at an angle.  Once more, the lighting could
   not have effect on those areas with normal bump mapping.  You could use displacement
   mapping but that would dramatically increase the number of polygons you would need
   for simple surfaces.  Using parallax mapping could solve this altogether.
   
   Parallax mapping needs a decal map (well optional), a bump map, and a height map.
   This height map is what is used to shade a surface to give it the appearance of being
   a complex 3D mesh.  It also helps us alter the lighting calculation in a way to where
   we can see those areas under bricks as if the brick were actually there is 3D space.
   The height map could be a simple black and white version of the decal texture image.
   With that height map you could create the bump map.
   
   So to make a long stroy short the only difference between normal bump mapping
   and parallax mapping is the addition of a height image.  With this you can calculate
   new texture coordinates that are used when fetching the decal texture color and bump
   map color (well technically it is a normal, not a color).
*/


// Demo: Parallax Mapping.


#define WIN32_LEAN_AND_MEAN
#define VC_LEANMEAN

#include<d3dx9.h>


// Function Prototypes...
bool InitializeDirect3D(HWND hwnd, bool fullscreen);
bool InitializeObject();
void RenderScene();
void ShutdownDirect3D();

#define D3DFVF_D3DVertex (D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_TEX1 | \
                          D3DFVF_XYZ | D3DFVF_XYZ)


// DirectX objects.
LPDIRECT3D9 Direct3D_Object = NULL;
LPDIRECT3DDEVICE9 D3D_Device = NULL;

// Textures.
LPDIRECT3DTEXTURE9 Texture = NULL;
LPDIRECT3DTEXTURE9 bumpMap = NULL;
LPDIRECT3DTEXTURE9 parallaxMap = NULL;

// Shader and vertex decalaration.
LPD3DXEFFECT effectShader = NULL;
LPDIRECT3DVERTEXDECLARATION9 ShaderDeclaration = NULL;

// Shader declaration.
D3DVERTEXELEMENT9 SDec[] =
{
   {0, 0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0},
   {0, 12, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_NORMAL, 0},
	{0, 24, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0},
   {0, 32, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TANGENT, 0},
   {0, 44, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BINORMAL, 0},
   D3DDECL_END()
};

// Vertex stucture.
struct Vertex
{
   FLOAT x, y, z;
   FLOAT nx, ny, nz;
   FLOAT tu, tv;
   FLOAT sx, sy, sz;
   FLOAT bx, by, bz;
};

// Our model object.
LPDIRECT3DVERTEXBUFFER9 Vertex_Buffer = NULL;

// Direction of our main light.
D3DXVECTOR4 lightPos = D3DXVECTOR4(0.0f, 0.0f, -11.0f, 0.0f);

// Our main 3 matrices...
D3DXMATRIX WorldMatrix;
D3DXMATRIX ViewMatrix;
D3DXMATRIX ProjectionMatrix;

// gWidth and gHeight is used for the aspect ratio.
int gWidth = 0;
int gHeight = 0;

// Mouse rotations.
float xRotation = 0.0f;
float yRotation = 0.0f;


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   static POINT oldMousePos;     // Last mouse position.
   static POINT currentMousePos; // Current mouse position.
   static bool  isMouseActive;   // Is the left mouse button down.
   int width, height;

   switch(message)
      {
         case WM_KEYUP:
            // If the user presses the escape key then exit the application.
            if(wParam == VK_ESCAPE)
               PostQuitMessage(0);
            break;

         case WM_CLOSE:
         case WM_DESTROY:
            PostQuitMessage(0);
            break;
         
         case WM_LBUTTONDOWN:
				oldMousePos.x = currentMousePos.x = LOWORD (lParam);
				oldMousePos.y = currentMousePos.y = HIWORD (lParam);
            isMouseActive = true;
			   break;


			case WM_LBUTTONUP:
				isMouseActive = false;
				break;

			case WM_MOUSEMOVE:
				currentMousePos.x = LOWORD (lParam);
				currentMousePos.y = HIWORD (lParam);

				if(isMouseActive)
					{
						xRotation -= (currentMousePos.x - oldMousePos.x);
						yRotation -= (currentMousePos.y - oldMousePos.y);
					}

				oldMousePos.x = currentMousePos.x;
				oldMousePos.y = currentMousePos.y;
			   break;

         case WM_SIZE:
            height = HIWORD(lParam);
            width = LOWORD(lParam);

            if(height == 0) height = 1;
            if(width == 0) width = 1;

            gWidth = width;
            gHeight = height;
            break;

         default:
         break;
      }

    return DefWindowProc(hwnd, message, wParam, lParam);
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                           // Message object.
   HWND hwnd;                                         // Handle to the window.
   WNDCLASSEX windowClass;                            // Window class object.
   bool done = false;                                 // True then exit.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);           // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;       // style of the window.
   windowClass.lpfnWndProc = WndProc;                 // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                        // Extra class information.
   windowClass.cbWndExtra = 0;                        // Extra window information.
   windowClass.hInstance = hInstance;                 // Handle of application Instance.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);// Window Icon.
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW); // Load mouse cursor.
   windowClass.hbrBackground = NULL;                  // Background color.
   windowClass.lpszMenuName = NULL;                   // Menu.
   windowClass.lpszClassName = "UGPClass";            // Name of the window class.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);// Minimized window icon.

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // Create the window.
   hwnd = CreateWindowEx(NULL,                              // The extended style.
                         "UGPClass",                        // Window class.
                         "Parallax Mapping Demo - by The Programming Ace",// Window name.
                          WS_OVERLAPPEDWINDOW | WS_VISIBLE |// Window style.
                          WS_SYSMENU |WS_CLIPCHILDREN |     // Window style.
                          WS_CLIPSIBLINGS,                  // Window style.
                          100, 100,                         // X, Y coords.
                          640, 480,                         // Window size.
                          NULL,                             // Handle to parent window.
                          NULL,                             // Menu.
                          hInstance,                        // Handle to app instance.
                          NULL);                            // Pointer to window.

   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW);    // Show the window.
   UpdateWindow(hwnd);           // Update its display.

   done = false;                 // false = run program, true means stop.

   // Initialize Direct3D.  If fails then we don't want to run the application.
   if(!InitializeDirect3D(hwnd, false)) done = true;

   // Application loop.
   while(!done)
      {
         if(PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     done = true;
                  }

               TranslateMessage(&msg);
               DispatchMessage(&msg);
            }
         else
            {
               RenderScene();
            }
      }

    // Here we shutdown Direct 3D.
	ShutdownDirect3D();

   // Here we unregister the window class with the OS.
   UnregisterClass("UGPClass", windowClass.hInstance);

   return (int)msg.wParam;
}


bool InitializeDirect3D(HWND hwnd, bool fullscreen)
{
   // This object will allow us to set the display mode of the screen.
   D3DDISPLAYMODE DisplayMode;

   // This will allow us to set the parameters of the screen.
   D3DPRESENT_PARAMETERS Present_Parameters;

   // This is used to get the capabilities of the hardware.
   D3DCAPS9 D3DCaps;

   // It is always a good idea to clear out memory in object although not necessary.
   ZeroMemory(&Present_Parameters, sizeof(Present_Parameters));

   // Create the Direct3D object to get everything started.
   Direct3D_Object = Direct3DCreate9(D3D_SDK_VERSION);

   // Error checking.  Make sure that it was successful.
   if(Direct3D_Object == NULL)
      {
         MessageBox(NULL, "Error, couldn't initialize DirectX!?!",
                    "Error!", MB_OK);
         return false;
      }

   // This function will get the display mode of the device and place it in DisplayMode.
   if(FAILED(Direct3D_Object->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &DisplayMode)))
      {
         MessageBox(NULL, "Error setting the display mode.", "Error!", MB_OK);
         return false;
      }

   // Get the capabilities of the hardware.
	if(FAILED(Direct3D_Object->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &D3DCaps)))
		return false;

   // Test which is supported, hardware or software vertex processing.
	DWORD VertexProcessing = 0;

	if(D3DCaps.VertexProcessingCaps != 0)
		VertexProcessing |= D3DCREATE_HARDWARE_VERTEXPROCESSING;
	else
		VertexProcessing |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;

   // Here we are setting the applications parameters...
   if(fullscreen)
      {
         Present_Parameters.Windowed = FALSE;               // Window mode (fullscreen).
         Present_Parameters.BackBufferWidth = 640;
         Present_Parameters.BackBufferHeight = 480;
      }
   else
      Present_Parameters.Windowed = TRUE;                   // Window mode.
   Present_Parameters.SwapEffect = D3DSWAPEFFECT_DISCARD;
   Present_Parameters.BackBufferFormat = DisplayMode.Format;
   Present_Parameters.BackBufferCount = 1;
   Present_Parameters.EnableAutoDepthStencil = TRUE;
   Present_Parameters.AutoDepthStencilFormat = D3DFMT_D16;

   // Now we must create the rendering device.
   if(FAILED(Direct3D_Object->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hwnd,
                                           D3DCREATE_SOFTWARE_VERTEXPROCESSING,
                                           &Present_Parameters, &D3D_Device)))
      {
         MessageBox(NULL, "CreateDevice() failed!  Make sure you have DirectX 9.",
                    "Error!", MB_OK);
         return false;
      }

   // One last check to be sure.
   if(D3D_Device == NULL)
      {
         MessageBox(NULL, "D3D_Device is equal to NULL!?!", "Error!", MB_OK);
         return false;
      }

   // Set our states.
   D3D_Device->SetRenderState(D3DRS_LIGHTING, FALSE);
	D3D_Device->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
   D3D_Device->SetRenderState(D3DRS_ZENABLE, TRUE);

   // We can use this to store any errors when compiling the shader effect file.
	LPD3DXBUFFER errors = NULL;

   // Load the effect file.
	if(FAILED(D3DXCreateEffectFromFile(D3D_Device, "ParallaxMapping.fx", NULL, NULL, 0,
                                      NULL, &effectShader, &errors)))
      {
         // Display errors.
		   LPVOID compileErrors = errors->GetBufferPointer();
		   
         // Show the errors to the user.
         MessageBox(NULL, (const char*)compileErrors, "Shader Compile Error...",
                    MB_OK | MB_ICONEXCLAMATION);

         return false;
	   }

   // Initialize out .x model.
   if(!InitializeObject()) return false;

   // Set the projection matrix.
   D3DXMatrixPerspectiveFovLH(&ProjectionMatrix, 0.8f,
                              (float)gWidth/(float)gHeight, 0.1f, 1000.0f);
   D3D_Device->SetTransform(D3DTS_PROJECTION, &ProjectionMatrix);

   return true;
}


bool InitializeObject()
{
   // Square data for its position and texture coords.
	Vertex Polygon_Data[4] =
   {
		{-5.0f, 5.0f, 1.0f, 0, 0, -1, 0.0, 0.0, 1, 0, 0, 0, -1, 0},
      {5.0f, 5.0f, 1.0f, 0, 0, -1, 1.0, 0.0, 1, 0, 0, 0, -1, 0},
	   {-5.0f, -5.0f, 1.0f, 0, 0, -1, 0.0, 1.0, 1, 0, 0, 0, -1, 0},
      {5.0f, -5.0f, 1.0f, 0, 0, -1, 1.0, 1.0, 1, 0, 0, 0, -1, 0}
	};

   // Create the vertex buffer that will hold the polygon.
	if(FAILED(D3D_Device->CreateVertexBuffer(sizeof(Polygon_Data),
                                            0, D3DFVF_D3DVertex,
                                            D3DPOOL_DEFAULT, &Vertex_Buffer, NULL)))
      {
         return false;
      }
	
   // Pointer to the buffer.
   Vertex* Vertices;

   // Lock the buffer we can write to it.
   if(FAILED(Vertex_Buffer->Lock(0, sizeof(Polygon_Data), (void**)&Vertices, 0)))
      return false;

   // Here we copy the polygon's data into the vertex buffer.
   memcpy(Vertices, Polygon_Data, sizeof(Polygon_Data));

   // Unlock when your done coping data into the buffer.
   Vertex_Buffer->Unlock();


   // Load the decal image.
   if(D3DXCreateTextureFromFile(D3D_Device, "decal.tga", &Texture) != D3D_OK)
      return false;

   // Load bump map.
   if(D3DXCreateTextureFromFile(D3D_Device, "normal.tga", &bumpMap) != D3D_OK)
      return false;
   
   // Load height map.
   if(D3DXCreateTextureFromFile(D3D_Device, "height.tga", &parallaxMap) != D3D_OK)
      return false;

   // Create the vertex declaration.
   if(D3D_Device->CreateVertexDeclaration(SDec, &ShaderDeclaration) != D3D_OK)
      return false;

   return true;
}


void SetShader()
{
   // Create a view matrix (set the camera).
   D3DXVECTOR3 pos(0.0f, 0.0f, -10.0f);
   D3DXVECTOR3 lookat(0.0f, 0.0f, 0.0f);
   D3DXVECTOR3 up(0.0f, 1.0f, 0.0f);
   D3DXMatrixLookAtLH(&ViewMatrix, &pos, &lookat, &up);
   
   // Set the rotation by mouse.
   D3DXMatrixRotationYawPitchRoll(&WorldMatrix, D3DXToRadian(xRotation),
                                    D3DXToRadian(yRotation), 0.0f);

   // Set the model view projection matrix.
   D3DXMATRIX modelViewProj = WorldMatrix * ViewMatrix * ProjectionMatrix;

   // Set the shader parameters.
   D3DXMatrixTranspose(&modelViewProj, &modelViewProj);
   effectShader->SetMatrix("worldViewProj", &modelViewProj);
   
   // Set the world matrix.
   effectShader->SetMatrix("worldView", &WorldMatrix);

   // Set the light position.
   effectShader->SetVector("lightPos", &lightPos);
}


void RenderScene()
{
   // Clear the screen.
   D3D_Device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                     D3DCOLOR_XRGB(0,0,0), 1.0f, 0);
    
   // Tell Direct3D we are now going to start drawing.
   D3D_Device->BeginScene();

      // Apply the textures.
      effectShader->SetTechnique("ParallaxMapping");
      effectShader->SetTexture("decal", Texture);
      effectShader->SetTexture("bump", bumpMap);
      effectShader->SetTexture("height", parallaxMap);

      // Set the vertex shader variables.
      SetShader();
      
      // Get total passes.
      effectShader->Begin(NULL, 0);

         effectShader->BeginPass(0);

            // Draw the square.
            D3D_Device->SetVertexDeclaration(ShaderDeclaration);
            D3D_Device->SetStreamSource(0, Vertex_Buffer, 0, sizeof(Vertex));
            D3D_Device->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);

         effectShader->EndPass();

      // End of the effect.
      effectShader->End();

   // We are done drawing for this scene.
   D3D_Device->EndScene();
    
   // Swap the old frame with the new one.
   D3D_Device->Present(NULL, NULL, NULL, NULL);
}


void ShutdownDirect3D()
{
   // Here we release the Direct3D objects and vertex shader.
   if(Texture != NULL) 
      {
         Texture->Release();
         Texture = NULL;
      }
   
   if(bumpMap != NULL) 
      {
         bumpMap->Release();
         bumpMap = NULL;
      }
   
   if(parallaxMap != NULL) 
      {
         parallaxMap->Release();
         parallaxMap = NULL;
      }

   if(effectShader != NULL)
      {
         effectShader->Release();
         effectShader = NULL;
      }

   if(ShaderDeclaration != NULL)
      {
         ShaderDeclaration->Release();
         ShaderDeclaration = NULL;
      }

   if(D3D_Device != NULL)
      {
         D3D_Device->Release();
         D3D_Device = NULL;
      }

   if(Direct3D_Object != NULL)
      {
         Direct3D_Object->Release();
         Direct3D_Object = NULL;
      }
   
   if(Vertex_Buffer != NULL)
      {
	      Vertex_Buffer->Release();
         Vertex_Buffer = NULL;
      }
}


// Recap:

// If you rotate the camera and move the light source you can swear that you are looking
// at a very complex 3D mesh, not a simulation.  With the help of the height map texture
// such detail is possible and adds a even higher level of realism than just bump mapping
// alone.  Just like how texture maps where created to add detail to surfaces, and bump maps
// where created to add a even higher level of detail to simple surfaces, parallax mapping
// can take things to a even higher level.


// Copyright October 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com