//*********************************************************
// www.Ultimategameprogramming.com
// Parallax Mapping shader demo.
// Created by the Programming Ace.
//
// Uniform Variables:
//   worldViewProj   - Holds the model view projection matrix.
//   worldView       - Holds the model view matrix.
//   lightPos        - Holds object space light position.
//   decal           - Holds the texture.
//   bump            - Holds bump map texture.
//   height          - Holds height map texture.
//
//*********************************************************


// Uniform parameters.
float4x4 worldViewProj  : WorldViewProjection;
float4x4 worldView      : WorldView;

float3 lightPos;

texture decal;
texture height;
texture bump;

// This sampler struct is used to specifiy texture options.
sampler DecalSampler = sampler_state
{
    Texture   = (decal);
    MipFilter = LINEAR;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};


sampler ParallaxSampler = sampler_state
{
    Texture   = (height);
    MipFilter = POINT;
    MinFilter = POINT;
    MagFilter = POINT;
};

sampler BumpSampler = sampler_state
{
    Texture   = (bump);
    MipFilter = LINEAR;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};


// Vertex shader input.
struct VS_Input
{
   float3 vertexPos  : POSITION;
   float3 normal     : NORMAL;
   float2 texture0   : TEXCOORD0;
   float3 sTangent   : TANGENT;
   float3 binormal   : BINORMAL;
};


// Vertex shader output.
struct Vs_Output
{
   float4 vertexPos  : POSITION;
   float2 texture0   : TEXCOORD0;
   float3 lightDirection : TEXCOORD1;
   float3 viewDirection : TEXCOORD2;
};


// Pixel shader output.
struct Ps_Output
{
	float4 color      : COLOR;
};


// The vertex shader...
Vs_Output VertexShaderEffect(VS_Input IN)
{
   Vs_Output vs_out;
   
   // Hardcode camera pos to save on uniform variables.  The camera
   // does not move so its not important.
   float3 camPos = float3(0, 0, -5.0);

   // Set output vertex pos, and texture coords.
   vs_out.vertexPos = mul(worldViewProj, float4(IN.vertexPos, 1));
   vs_out.texture0 = IN.texture0;
   
   // Calculate world space position.
   float4 pos = mul(worldView, float4(IN.vertexPos, 1));
   
   // Build TBN matrix.
   float3x3 tbnMatrix;
   tbnMatrix[0] = mul(IN.sTangent, worldView);
   tbnMatrix[1] = mul(IN.binormal, worldView);
   tbnMatrix[2] = mul(IN.normal, worldView);
   
   // Set the light direction then convert to texture space.
   vs_out.lightDirection = lightPos - IN.vertexPos;
   vs_out.lightDirection = mul(tbnMatrix, vs_out.lightDirection);
   
   // Set the view direction then convert to texture space.
   vs_out.viewDirection = camPos - pos;
   vs_out.viewDirection = mul(tbnMatrix, vs_out.viewDirection);
   
   return vs_out;
}


// The pixel shader...
Ps_Output PixelShaderEffect(Vs_Output IN)
{
   Ps_Output ps_out;
   
   float scale = 0.04f;
   float bias = 0.02;
   
   float4 heightCol = tex2D(ParallaxSampler, IN.texture0);

   // Normalize the light direction on a per pixel level.
   float3 lightDirection = normalize(IN.lightDirection);
   float3 viewDirection = normalize(IN.viewDirection);
   
   float height = scale * heightCol - bias;
   float2 newTexCoord = height * viewDirection + IN.texture0;
   
   float4 decalCol = tex2D(DecalSampler, newTexCoord);
   float4 bumpCol = tex2D(BumpSampler, newTexCoord);
   
   // Expand normal.
   bumpCol = (bumpCol - 0.5f) * 2;
   
   // Calculate the bump value as n dot l.
   float diffuse = saturate(dot(bumpCol, lightDirection));
  
   // To make things simple I hardcoded material values but we could have passed
   // these as uniform parameters.
   float3 diffuseMat = float3(0.8f, 0.8f, 0.8f);
   float ambient = 0.4f;
   
   // Combine the decal with the bump value.
   ps_out.color = decalCol * float4((ambient + (diffuseMat * diffuse)), 1);

   return ps_out;
}


// Effect...
technique ParallaxMapping
{
   pass Pass0
   {
      // Here we will compile and set the vertex and pixel shader.
      VertexShader = compile vs_2_0 VertexShaderEffect();
      PixelShader  = compile ps_2_0 PixelShaderEffect();
   }
}